import functools
import inspect


def vba_func(func):
    @functools.wraps(func)
    def inner(func):
        if not hasattr(func, '__vba__'):
            func.__vba__ = describe_function(func)
        func.__vba__['sub'] = False
        return func
    return inner(func)


def vba_sub(func):
    @functools.wraps(func)
    def inner(func):
        if not hasattr(func, '__vba__'):
            func.__vba__ = describe_function(func)
        func.__vba__['sub'] = True
        return func
    return inner(func)


def vba_object(func):
    @functools.wraps(func)
    def inner(func):
        func.__vba__['object'] = True
        return func
    return inner(func)


def describe_function(func):
    info = {}
    info['name'] = func.__name__
    info['sub'] = False
    info['object'] = False
    (args, argsmap) = _describe_function_args(func)
    info['args'] = args
    info['argsmap'] = argsmap
    return info


def extract_function_description(func):
    if hasattr(func, '__vba__'):
        return func.__vba__
    else:
        return describe_function(func)


def _describe_function_args(func):
    func_sig = _extract_function_args(func)
    nArgs = len(func_sig['args'])
    nDefaults = len(func_sig['defaults'])
    nRequiredArgs = nArgs - nDefaults
    if func_sig['vararg'] and nDefaults > 0:
        raise Exception("Not supported: function with both optional and variable length arguments")

    args = []
    argsmap = {}
    for (vpos, vname) in enumerate(func_sig['args']):
        arg_info = {
            'name': vname,
            'pos': vpos,
            'vararg': vname == func_sig['vararg'],
            'options': {}
        }
        if vpos >= nRequiredArgs:
            arg_info['optional'] = func_sig['defaults'][vpos - nRequiredArgs]
        args.append(arg_info)
        argsmap[vname] = args[-1]

    return (args, argsmap)


def _extract_function_args(func):
    signature = inspect.signature(func)
    vararg = None
    args = []
    defaults = []
    for arg in signature.parameters.values():
        if arg.kind is inspect.Parameter.POSITIONAL_OR_KEYWORD:
            args.append(arg.name)
            if arg.default is not inspect.Signature.empty:
                defaults.append(arg.default)
        elif arg.kind is inspect.Parameter.VAR_POSITIONAL:
            args.append(arg.name)
            vararg = arg.name
        else:
            raise Exception("Not supported: keyword arguments")
    return {
        'args': args,
        'defaults': defaults,
        'vararg': vararg
    }
