import sys
import os

from importlib import import_module, reload

from .writer import VBAWriter
from .decorators import extract_function_description


def reload_module(module_name: str):
    reload(_get_module(module_name))


def preload_module(module_path: str) -> str:
    (folder, file) = os.path.split(module_path)
    if folder not in sys.path:
        sys.path.append(folder)
    module_name = os.path.splitext(file)[0]
    _get_module(module_name)
    return module_name


def call_function(module_name, func_name, args=None):
    module = _get_module(module_name)
    func = getattr(module, func_name)
    if args is not None:
        result = func(*args)
    else:
        result = func()
    return result


def wrap_python(source: str, destination: str, encoding: str) -> str:
    source_module = preload_module(source)

    out_file = open(destination, 'w+', encoding=encoding)
    (folder, file) = os.path.split(destination)

    vba = VBAWriter(out_file)
    vba.line(f'Attribute VB_Name = "{os.path.splitext(file)[0]}"')
    vba.line('Option Private Module')
    vba.line('Option Explicit')
    vba.line("'Generated code by python_api - changes will be lost with next generation!")

    _wrap_module(source_module, vba)

    out_file.close()
    return destination


def _wrap_module(module_name: str, vba: VBAWriter):
    module = _get_module(module_name)
    for attribute_item in map(lambda attr: getattr(module, attr), dir(module)):
        if not hasattr(attribute_item, '__vba__'):
            continue
        info = extract_function_description(attribute_item)
        _wrap_function(module_name, info, vba)


def _wrap_function(module_name, info, vba: VBAWriter):
    fname = f"Py_{info['name']}"
    ftype = 'Sub' if info['sub'] else 'Function'
    (func_sig, vararg) = _generate_function_signature(ftype, fname, info)
    args_vba = _wrap_vararg(vararg, info['args'], vba)

    vba.line('')
    vba.start_block(func_sig)

    vba.line("Dim iPython as API_Python: Set iPython = AccessPython")
    vba.line("On Error GoTo FAILED_CALL")
    if info['object']:
        vba.line(f"""Set {fname} = iPython.CallFunctionReturnObject("{module_name}", "{info['name']}", {args_vba})""")
    else:
        vba.line(f"""{fname} = iPython.CallFunction("{module_name}", "{info['name']}", {args_vba})""")
    vba.line("On Error GoTo 0")
    vba.line(f"Exit {ftype}")
    vba.label("FAILED_CALL")
    vba.line(f"{fname} = Err.Description")

    vba.end_block('End ' + ftype)


def _wrap_vararg(vararg, args, vba: VBAWriter) -> str:
    if vararg == '':
        return 'Array(' + ', '.join(arg['name'] for arg in args) + ')'

    vba.line("Dim argsArray() As Variant")
    non_varargs = [arg['name'] for arg in args if not arg['vararg']]
    vba.line("argsArray = Array(%s)" % tuple({', '.join(non_varargs)}))
    vba.line(f"ReDim Preserve argsArray(0 to UBound({vararg}) - LBound({vararg}) + {len(non_varargs)})")
    vba.line("Dim k&")
    vba.line(f"For k = LBound({vararg}) To UBound({vararg})")
    vba.line(f"argsArray({len(non_varargs)} + k - LBound({vararg})) = {vararg}(k)")
    vba.line("Next k")
    return 'argsArray'


def _generate_function_signature(ftype: str, fname: str, info):
    args = ''
    first = True
    vararg = ''
    for arg in info['args']:
        argname = f"pyArg_{arg['name']}"
        arg['name'] = argname
        if not first:
            args += ', '
        if 'optional' in arg:
            args += 'Optional '
        elif arg['vararg']:
            args += 'ParamArray '
            vararg = argname
        args += argname
        if arg['vararg']:
            args += '()'
        first = False
    return (f'Public {ftype} {fname}({args})', vararg)


# Imported module caching
_loaded_modules = {}


def _get_module(module_name):
    module_info = _loaded_modules.get(module_name, None)
    if module_info is not None:
        module = module_info['module']
    else:
        module = import_module(module_name)
        filename = os.path.normcase(module.__file__.lower())
        _loaded_modules[module_name] = {'filename': filename, 'module': module}
    return module