#include "pch.h"

namespace vbatopy {

Server::Server() {
	if (FAILED(::CoCreateGuid(&clsid))) {
		throw std::runtime_error("CoCreateGuid failed.");
	}
}

Server::~Server() {
	this->ForceKill();
}

Server& Server::Instance() {
	static Server py{};
	return py;
}

void Server::Setup(std::string command, const bool showConsole_) {
	this->Kill();
	showConsole = showConsole_;
	execCommand = std::move(command);
	SubstituteMacro(execCommand);
}

bool Server::CheckStatus() const {
	if (dispatcher == nullptr) {
		return false;
	}
	UINT info{};
	return dispatcher->GetTypeInfoCount(&info) == S_OK;
}

void Server::Kill() {
	ReleaseDispatch();
	ReleaseKillerJob();
	pyThread = nullptr;
	pyProcess = nullptr;
}

void Server::ForceKill() {
	ReleaseKillerJob();
	dispatcher = nullptr;
	pyThread = nullptr;
	pyProcess = nullptr;
}

void Server::Start() {
	CreateInternalProcess();
	utils::AutoCloseHandle hPiThread(pyThread);
	utils::AutoCloseHandle hPiProcess(pyProcess);

	CreateKillerJob();
	utils::AutoCloseHandle hJobAuto(jobHandle);

	ActivateDispatcher();

	jobHandle = hJobAuto.Release();
}

void Server::ReleaseDispatch() {
	if (dispatcher != nullptr) {
		dispatcher->Release();
		dispatcher = nullptr;
	}
}

void Server::ReleaseKillerJob() {
	if (jobHandle != nullptr) {
		::CloseHandle(jobHandle);
		jobHandle = nullptr;
	}
}

void Server::SubstituteMacro(std::string& path) const {
	static const auto MACRO_LEN = 8U;
	const auto position = path.find("$(CLSID)", 0);
	if (position ==	std::string::npos) {
		return;
	}
	path = path.substr(0, position) + utils::GUIDToStr(clsid) + path.substr(position + MACRO_LEN, path.length() - position - MACRO_LEN);
}

void Server::CreateInternalProcess() {
	static constexpr auto CMD_INPUT_MAX_LENGTH = 8000U;

	const DWORD windowFlags = showConsole ? NORMAL_PRIORITY_CLASS | CREATE_BREAKAWAY_FROM_JOB :
		NORMAL_PRIORITY_CLASS | CREATE_BREAKAWAY_FROM_JOB  | CREATE_NO_WINDOW;

	if (execCommand.length() > CMD_INPUT_MAX_LENGTH) {
		throw std::length_error("Start server command is too long");
	}

	char cmdLine[CMD_INPUT_MAX_LENGTH] = "";
	strcat_s(cmdLine, execCommand.c_str());

	STARTUPINFOA si;
	ZeroMemory(&si, sizeof(si));
	si.cb = sizeof(si);
	PROCESS_INFORMATION pi;
	ZeroMemory(&pi, sizeof(pi));

	if (!::CreateProcessA(
		nullptr, 
		cmdLine,  // NOLINT(cppcoreguidelines-pro-bounds-array-to-pointer-decay, hicpp-no-array-decay)
		nullptr, nullptr, TRUE, windowFlags, nullptr, nullptr, 
		&si, &pi)) {
		throw std::runtime_error(std::format("Could not create Python: {}\nCommand{}", utils::GetLastErrorMessage(), execCommand));
	}
	pyProcess = pi.hProcess;
	pyThread = pi.hThread;
}

void Server::CreateKillerJob() {
	utils::AutoCloseHandle hJobAuto{ CreateJobObject(nullptr, nullptr) };
	if (hJobAuto.handle == nullptr) {
		throw std::runtime_error("CreateJobObject failed: " + utils::GetLastErrorMessage());
	}
	JOBOBJECT_EXTENDED_LIMIT_INFORMATION jxli;
	ZeroMemory(&jxli, sizeof(jxli));
	jxli.BasicLimitInformation.LimitFlags = JOB_OBJECT_LIMIT_KILL_ON_JOB_CLOSE;
	if (!::SetInformationJobObject(hJobAuto.handle, JobObjectExtendedLimitInformation, &jxli, sizeof(jxli))) {
		throw std::runtime_error("SetInformationJobObject failed: " + utils::GetLastErrorMessage());
	}
	if (!::AssignProcessToJobObject(hJobAuto.handle, pyProcess)) {
		throw std::runtime_error("AssignProcessToJobObject failed: " + utils::GetLastErrorMessage());
	}
	jobHandle = hJobAuto.Release();
}

void Server::ActivateDispatcher() {
	static constexpr auto RETRY_COUNT = 600U;
	static constexpr auto RETRY_PERIOD_MS = 100U;
	for (uint16_t attempt = 0; attempt < RETRY_COUNT; ++attempt) {
		if (::CoCreateInstance(clsid, nullptr, CLSCTX_LOCAL_SERVER, IID_IDispatch, reinterpret_cast<void**>(&(dispatcher))) != REGDB_E_CLASSNOTREG) {
			return;
		}
		DWORD dwExitCode{};
		if (!::GetExitCodeProcess(pyProcess, &dwExitCode)) {
			throw std::runtime_error("GetExitCodeProcess failed: " + utils::GetLastErrorMessage());
		}
		if (dwExitCode != STILL_ACTIVE) {
			throw std::runtime_error("Cmd process exited before it was possible to create the interface object.\nCommand: " + execCommand);
		}
		::Sleep(RETRY_PERIOD_MS);
	}
	throw std::runtime_error("Could not activate Python COM server");
}

} // namespace vbatopy